(* SDL.getticks () -- how many hundredths of a second since SDL started? *)
(* event loop: a case statement *)

val screen = SDL.makescreen (700, 700)

fun requireimage s =
      let val s = "testgraphics/" ^ s
      in
          case SDL.Image.load s of
              NONE => (print ("couldn't open " ^ s ^ "\n");
                       raise Fail "end of death")
            | SOME p => p
      end

  structure Font = FontFn (val surf = requireimage "font.png"
                           val charmap =
                           " ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789" ^
                           "`-=[]\\;',./~!@#$%^&*()_+{}|:\"<>?" (* \" *)
                           (* CHECKMARK ESC HEART LCMARK1 LCMARK2 BAR_0 BAR_1 BAR_2 BAR_3-
                              BAR_4 BAR_5 BAR_6 BAR_7 BAR_8 BAR_9 BAR_10 BARSTART LRARROW LLARROW *)
                           val width = 9
                           val height = 16
                           val styles = 6
                           val overlap = 1
                           val dims = 3)

val black = SDL.color (0w0, 0w0, 0w0, 0w0)
fun spew s = (SDL.clearsurface (screen, black);
              Font.draw (screen, 50, 42, s);
              SDL.flip screen)

fun status s = Font.draw (screen, 10, 10, s)

val starfish = requireimage "redstar.png"

fun win () =
    (spew "win!!!!  [press any key]";
     let fun l () =
        case SDL.pollevent () of
            NONE => (SDL.delay 0; l ())
          | SOME (SDL.E_KeyDown _) => ()
          | SOME (SDL.E_Quit) => ()
          | SOME _ => l ()
     in
        l ()
     end)

(*
fun render (state as {x, y, ...}) =
    (SDL.clearsurface (screen, black);
     status ("Time: " ^ Word32.toString (SDL.getticks ()));
     SDL.blitall (starfish, screen, x, y);
     SDL.flip screen)

val init_state = { x = 120, y = 40, dx = 0, dy = 0 }

(* bouncecrab game! *)
fun update (state as {x, y, dx, dy}) =
    let fun inbounds_x x = x >= 0 andalso x < SDL.surface_width screen
        fun inbounds_y y = y >= 0 andalso y < SDL.surface_height screen
        val newx = x + dx
        val newy = y + dy
        val (x, dx) = if inbounds_x newx then (newx, dx) else (x, ~dx)
        val (y, dy) = if inbounds_y newy then (newy, dy) else (y, ~dy)
    in
        {x=x, y=y, dx=dx, dy=dy}
    end
*)

fun render (state as {t, x, y, vx, vy, ax, ay}) =
    let fun r2s r = StringCvt.padLeft #"0" 9 (Real.fmt (StringCvt.FIX (SOME 2)) r)
    in
    (SDL.clearsurface (screen, black);
     status ("x=" ^ r2s x ^ ",y=" ^ r2s y
          ^ ",vx=" ^ r2s vx ^ ",vy=" ^ r2s vy
          ^ ",ax=" ^ r2s ax ^ ",ay=" ^ r2s ay);
     SDL.blitall (starfish, screen, Real.round x, Real.round y);
     SDL.flip screen)
    end

val init_state = { t = 0w0, x = 320.0, y = 320.0, vx = 0.0, vy = 0.0, ax = 0.0, ay = 0.0 }

fun update (state as {t=t0, x, y, vx, vy, ax, ay}) =
    let fun r x = Real.round x
        fun inbounds_x x = r x >= 0 andalso r x < SDL.surface_width screen
        fun inbounds_y y = r y >= 0 andalso r y < SDL.surface_height screen
        val t = SDL.getticks ()
        val dt = Real.fromInt (Word32.toInt (t - t0))
        val newx = x + vx * dt + 0.5 * ax * Math.pow (dt, 2.0)
        val newy = y + vy * dt + 0.5 * ay * Math.pow (dt, 2.0)
        val vx = vx + ax * dt
        val vy = vy + ay * dt
        val dv = 0.01
        fun dec v = if Real.abs v < dv then 0.0 else v - real(Real.sign v) * dv
        val vx = dec vx
        val vy = dec vy
        val (x, vx) = if inbounds_x newx then (newx, vx) else (x, ~vx)
        val (y, vy) = if inbounds_y newy then (newy, vy) else (y, ~vy)
    in
        {t=t, x=x, y=y, vx=vx, vy=vy, ax=0.0, ay=0.0}
    end

fun loop state =
    let
        val state as {t, x, y, vx, vy, ax, ay} = update state
        val () = render state
        val da = 0.01
    in
    case SDL.pollevent () of
        NONE => (SDL.delay 0; loop state)
      (* quit on escape or Mac quit *)
      | SOME (SDL.E_KeyDown { sym = SDL.SDLK_ESCAPE }) => ()
      | SOME SDL.E_Quit => ()
      (* win on w! *)
      | SOME (SDL.E_KeyDown { sym = SDL.SDLK_w }) => win ()
      (* change dx and dy on arrow keys *)
      | SOME (SDL.E_KeyDown { sym = SDL.SDLK_LEFT }) =>
            loop { t=t, x=x, y=y, vx=vx, vy=vy, ax= ~da, ay=0.0 }
      | SOME (SDL.E_KeyDown { sym = SDL.SDLK_RIGHT }) =>
            loop { t=t, x=x, y=y, vx=vx, vy=vy, ax=da, ay=0.0 }
      | SOME (SDL.E_KeyDown { sym = SDL.SDLK_UP }) =>
            loop { t=t, x=x, y=y, vx=vx, vy=vy, ax=0.0, ay= ~da }
      | SOME (SDL.E_KeyDown { sym = SDL.SDLK_DOWN }) =>
            loop { t=t, x=x, y=y, vx=vx, vy=vy, ax=0.0, ay=da }
      | SOME _ => loop state
    end

val () =
    let in
      spew "test";
      loop init_state
    end
